"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const parse5 = require("parse5");
const treeAdapter = require("parse5-htmlparser2-tree-adapter");
/**
 * Analyzes a given template expression for traversing its contained
 * HTML tree.
 */
class TemplateAnalyzer {
    /**
     * Constructor
     *
     * @param {ESTree.TaggedTemplateExpression} node Node to analyze
     */
    constructor(node) {
        this._node = node;
        let html = '';
        for (let i = 0; i < node.quasi.quasis.length; i++) {
            const quasi = node.quasi.quasis[i];
            const expr = node.quasi.expressions[i];
            html += quasi.value.raw;
            if (expr) {
                html += this._getExpressionPlaceholder(node, quasi);
            }
        }
        this._ast =
            parse5.parseFragment(html, {
                treeAdapter: treeAdapter,
                sourceCodeLocationInfo: true
            });
    }
    /**
     * Returns the ESTree location equivalent of a given parsed location.
     *
     * @param {treeAdapter.Node} node Node to retrieve location of
     * @return {?ESTree.SourceLocation}
     */
    getLocationFor(node) {
        if (treeAdapter.isElementNode(node)) {
            const loc = node.sourceCodeLocation;
            if (loc) {
                return this._resolveLocation(loc.startTag);
            }
        }
        else if (treeAdapter.isCommentNode(node) ||
            treeAdapter.isTextNode(node)) {
            const loc = node
                .sourceCodeLocation;
            if (loc) {
                return this._resolveLocation(loc);
            }
        }
        return this._node.loc;
    }
    /**
     * Returns the ESTree location equivalent of a given attribute
     *
     * @param {treeAdapter.Element} element Element which owns this attribute
     * @param {string} attr Attribute name to retrieve
     * @return {?ESTree.SourceLocation}
     */
    getLocationForAttribute(element, attr) {
        if (!element.sourceCodeLocation) {
            return null;
        }
        const loc = element.sourceCodeLocation.startTag.attrs[attr];
        return loc ? this._resolveLocation(loc) : null;
    }
    /**
     * Generates a placeholder string for a given quasi
     *
     * @param {ESTree.TaggedTemplateExpression} node Root node
     * @param {ESTree.TemplateElement} quasi Quasi to generate placeholder
     * for
     * @return {string}
     */
    _getExpressionPlaceholder(node, quasi) {
        const i = node.quasi.quasis.indexOf(quasi);
        if (/=$/.test(quasi.value.raw)) {
            return `"{{__Q:${i}__}}"`;
        }
        return `{{__Q:${i}__}}`;
    }
    /**
     * Resolves a Parse5 location into an ESTree location
     *
     * @param {parse5.Location} loc Location to convert
     * @return {ESTree.SourceLocation}
     */
    _resolveLocation(loc) {
        let offset = 0;
        for (const quasi of this._node.quasi.quasis) {
            const placeholder = this._getExpressionPlaceholder(this._node, quasi);
            offset += quasi.value.raw.length + placeholder.length;
            if (loc.startOffset < offset) {
                return quasi.loc;
            }
        }
        return null;
    }
    /**
     * Traverse the inner HTML tree with a given visitor
     *
     * @param {Visitor} visitor Visitor to apply
     * @return {void}
     */
    traverse(visitor) {
        const visit = (node, parent) => {
            if (!node) {
                return;
            }
            if (visitor.enter) {
                visitor.enter(node, parent);
            }
            if (node.type === 'root') {
                if (visitor.enterDocumentFragment) {
                    visitor.enterDocumentFragment(node, parent);
                }
            }
            else if (treeAdapter.isCommentNode(node)) {
                if (visitor.enterCommentNode) {
                    visitor.enterCommentNode(node, parent);
                }
            }
            else if (treeAdapter.isTextNode(node)) {
                if (visitor.enterTextNode) {
                    visitor.enterTextNode(node, parent);
                }
            }
            else if (treeAdapter.isElementNode(node)) {
                if (visitor.enterElement) {
                    visitor.enterElement(node, parent);
                }
            }
            if (treeAdapter.isElementNode(node) || node.type === 'root') {
                const children = node.childNodes;
                if (children && children.length > 0) {
                    children.forEach((child) => {
                        visit(child, node);
                    });
                }
            }
            if (visitor.exit) {
                visitor.exit(node, parent);
            }
        };
        visit(this._ast, null);
    }
}
exports.TemplateAnalyzer = TemplateAnalyzer;
